#include "scene.h"
#include "triangle.h"
#include "tmesh.h"

Scene *scene;

#include <iostream>

using namespace std;


Scene::Scene() {

  cgi = 0;
  soi = 0;

  cgi = new CGInterface();
  soi = new ShaderOneInterface();

  shadowMap = 0;
  fbmesh = 0;

  gui = new GUI();
  gui->show();

  int u0 = 20;
  int v0 = 50;
  int sci = 1;
  int w = sci*320;
  int h = sci*240;
  fb = new FrameBuffer(u0, v0, w, h);
  fb->label("SW framebuffer");
  fb->show();
  gui->uiw->position(fb->w+20+20, 50);

  hwfb = new FrameBuffer(u0, v0*2+h, w, h);
  hwfb->isHW = true;
  hwfb->label("HW framebuffer");
  hwfb->show();


  Setup();
  Render();

}

void Scene::Render() {


  if (true) {
    fb->FrameSetup(0xFFFFFFFF);
    V3 color(1.0f, 0.0f, 0.0f);
    tmesh.RenderFilled(color, ppc, fb, shadowMap, shadowPPC);

    if (!fbmesh) {
      fbmesh = new TMesh(fb, ppc);
    }

    fb->redraw();
  }

  hwfb->redraw();

}

void Scene::Setup() {

  float hfov = 45.0f;
  ppc = new PPC(fb->w, fb->h, hfov);

#if 1
  tmesh.Load("C:/Users/popescu/Desktop/tmp/geometry/teapot1k.bin");
//  tmesh.Load("C:/Users/popescu/Desktop/tmp/geometry/happy2.bin");
//  tmesh.Load("C:/Users/popescu/Desktop/tmp/geometry/auditorium.bin");
  V3 center = tmesh.GetCenter();
  cerr << "INFO: mesh center: " << center << endl;
  V3 newCenter(0.0f, 0.0f, -60.0f);
  tmesh.SetCenter(newCenter);
  float aabbDiagonal = 50.0f;
  tmesh.Scale(aabbDiagonal);
#endif

#if 0
  tmesh.SetRectangle(V3(0.0f, 0.0f, -200.0f), V3(1.0f, 0.0f, 0.0f),
    V3(0.0f, 1.0f, 0.0f), 40.0f, 40.0f);
  texture = new FrameBuffer(fb->w+gui->uiw->w() + 64, 50, 512, 512);
  texture->Set(0xFF0088FF);
  texture->DrawCircle(180.0f, 180.0f, 32, 0xFFFF0000);
  for (int i = 0; i < 7; i++) {
    texture->DrawSegment(V3((float)i, 0.0f, 0.0f), V3((float)i, 256.0f, 0.0f), 0xFF000000);
    texture->DrawSegment(V3(0.0f, (float)i, 0.0f), V3(256.0f, (float)i, 0.0f), 0xFF000000);
    texture->DrawSegment(V3(255.0f-(float)i, 0.0f, 0.0f), V3(255.0f-(float)i, 256.0f, 0.0f), 0xFF000000);
    texture->DrawSegment(V3(0.0f, 255.0f-(float)i, 0.0f), V3(256.0f, 255.0f-(float)i, 0.0f), 0xFF000000);
  }
  unsigned cols[2] = {0xFF000000, 0xFFFFFFFF};
  int cw = 2;
  texture->SetChecker(cols, cw);
  texture->label("texture");
  texture->show();
  tmesh.texture = texture;

  tmesh1.SetRectangle(V3(40.0f, 0.0f, -80.0f), V3(1.0f, 0.0f, 0.0f),
    V3(0.0f, 1.0f, 0.0f), 40.0f, 20.0f);
  texture1 = new FrameBuffer(fb->w+gui->uiw->w() + 64, 50, 256, 256);
  unsigned cols1[2] = {0xFF00FF00, 0xFFAA00FF};
  int cw1 = 32;
  texture1->SetChecker(cols1, cw1);
  texture1->label("texture1");
//  texture1->show();
  tmesh1.texture = texture1;

#endif

}

void Scene::DBG() {

  {
    shadowMap = new FrameBuffer(0, 0, fb->w, fb->h);
    for (int uv = 0; uv < fb->w*fb->h; uv++) {
      shadowMap->pix[uv] = fb->pix[uv];
      shadowMap->zb[uv] = fb->zb[uv];
    }
    shadowPPC = new PPC(*ppc);
    shadowMap->show();
    return;
  }

  {

    float hfov = 45.0f;
    PPC ppc(fb->w, fb->h, hfov);
    TMesh tmesh;
    tmesh.Load("C:/Users/popescu/Desktop/tmp/geometry/teapot1k.bin");
    V3 center = tmesh.GetCenter();
    cerr << "INFO: mesh center: " << center << endl;
    V3 newCenter(0.0f, 0.0f, -40.0f);
    tmesh.SetCenter(newCenter);
    float aabbDiagonal = 30.0f;
    tmesh.Scale(aabbDiagonal);
    V3 color(0.0f, 0.0f, 0.0f);

    while (1) {
      fb->FrameSetup(0xFFFFFFFF);
      V3 axisd = V3(1.0f, 1.0f, 1.0f).Normalized();
      tmesh.Spin(axisd, 1.0f);

      //    tmesh.RenderPoints(color, &ppc, fb);
      //    tmesh.RenderWireFrame(color, &ppc, fb);
      tmesh.RenderFilled(color, &ppc, fb, 0, 0);
      fb->redraw();
      Fl::check();
    }
    return;

  }

  {
    float hfov = 45.0f;
    PPC ppc(fb->w, fb->h, hfov);
    V3 P0(-20.0f, -40.0f, -390.0f);
    V3 P1(+20.0f, -20.0f, -170.0f);
    AABB aabb;
    aabb.AddPoint(P0);
    aabb.AddPoint(P1);
    TMesh tmesh;
    tmesh.Load("C:/Users/popescu/Desktop/tmp/geometry/teapot1k.bin");
//    tmesh.SetAABB(aabb);
    V3 color(0.0f, 0.0f, 0.0f);
//    tmesh.RenderPoints(color, &ppc, fb);
    tmesh.RenderWireFrame(color, &ppc, fb);
    fb->redraw();
    return;

  }

  {
    float hfov = 45.0f;
    PPC ppc0(fb->w, fb->h, hfov);
    PPC ppc1(fb->w, fb->h, hfov);
    ppc1.TranslateFW(50.0f);
    ppc1.Visualize(&ppc0, 10.0f, fb);
    fb->redraw();
    return;

  }

  {
    float hfov = 45.0f;
    PPC ppc(fb->w, fb->h, hfov);
    V3 P0(-20.0f, -10.0f, -250.0f);
    V3 P1(+20.0f, 10.0f, -50.0f);
    AABB aabb;
    aabb.AddPoint(P0);
    aabb.AddPoint(P1);
    V3 colors[2];
    colors[0] = V3(1.0f, 0.0f, 0.0f);
    colors[1] = V3(0.0f, 0.0f, 0.0f);
    fb->DrawAABB3D(aabb, colors, &ppc);
    fb->redraw();
    return;

  }

  {
    float hfov = 45.0f;
    PPC ppc(fb->w, fb->h, hfov);
    V3 P0(-10.0f, 0.0f, -100.0f);
    V3 P1(+10.0f, 0.0f, -100.0f);
    V3 c0(1.0f, 0.0f, 0.0f);
    V3 c1(0.0f, 1.0f, 0.0f);
    fb->DrawSegment3D(P0, P1, c0, c1, &ppc);
    fb->redraw();
    return;

  }


  {
    float hfov = 45.0f;
    PPC ppc(fb->w, fb->h, hfov);
    V3 P(0.0f, 0.0f, +100.0f), projP;
    bool projResult = ppc.Project(P, projP);
    return;

  }

  V3 verts[3] = {
    V3((float)fb->w/2.0f-10.0f, 13.0f, 0.0f),
    V3((float)17.0f, (float)fb->h/4.0f*3.0f, 0.0f),
    V3((float)fb->w-23.4, (float)fb->h/5.0f*4.0f, 0.0f)};
  Triangle tri;
  tri.verts[0] = verts[0];
  tri.verts[1] = verts[1];
  tri.verts[2] = verts[2];
  tri.color = V3(1.0f, 0.0f, 0.0f);
  tri.colors[0] = V3(1.0f, 0.0f, 0.0f);
  tri.colors[1] = V3(0.0f, 1.0f, 0.0f);
  tri.colors[2] = V3(0.0f, 0.0f, 1.0f);
//  tri.Rasterize(fb);
  tri.DrawWireFrame(fb);
  fb->redraw();
  return;

  fb->DrawCircle(120.4f, 103.4f, 50.0f, 0xFFFF0000);
  fb->redraw();
  return;

  fb->DrawCircle(120.4f, 103.4f, 50.0f, 0xFFFF0000);
  fb->DrawCircle(220.4f, 123.4f, 75.0f, 0xFF0000FF);
  FrameBuffer *edgeMap = fb->FindEdges();
  fb->redraw();
  edgeMap->position(fb->w+gui->uiw->w()+60, 50);
  edgeMap->show();
  return;


  V3 p0(14.1f, 20.2f, 0.0f);
  V3 p1(37.3f, 189.9f, 0.0f);
  fb->DrawSegment(p0, p1, 0xFF000000);
  fb->redraw();
  return;

  fb->DrawPoint(100, 200, 7, 0xFFFF0000);
  fb->redraw();
  return;

  int framesN = 1000;
  while (1) {
    for (int fi = 0; fi < framesN; fi++) {
      float currColorf = (float)fi/(float)(framesN-1);
      unsigned int currColor = 0xFF000000 + 
        (unsigned int) (255.0f*currColorf);
      fb->Set(currColor);
      fb->redraw();
      Fl::check();
    }
  }

}

void Scene::RenderGL() {

  glClearColor(0.0f, 0.0f, 0.5f, 1.0f);
  glClear(GL_COLOR_BUFFER_BIT |
    GL_DEPTH_BUFFER_BIT);
  // set view intrisics
  ppc->SetViewGL();
  // set view extrisics
  // draw geometry
  if (false && fbmesh) {
    fbmesh->RenderHW();
  }
  else {
    tmesh.RenderHW();
  }

}

void Scene::RenderGPU() {

  if (cgi->needInit) {
    cgi->PerSessionInit();
    soi->PerSessionInit(cgi);
  }

  glClearColor(0.0f, 0.0f, 0.5f, 1.0f);
  glClear(GL_COLOR_BUFFER_BIT |
    GL_DEPTH_BUFFER_BIT);

  ppc->SetViewGL();

  // per frame parameter setting and enabling programs
  soi->PerFrameInit();
  soi->BindPrograms();
  cgi->EnableProfiles();

  tmesh.RenderHW();

  cgi->DisableProfiles();

}


#if 0

for (tmi = 0; tmi < tmN; tmi++) {
  for (ti = 0; ti < tms[tmi].trisN; ti++) {
    Triangle currTri = tms[tmi].triangles[ti];
    // projection
    for (vi = 0; vi < 3; vi++) {
      if (!GPU) {
        projVertex[vi] = Project(currTri.vertex[vi], ppc);
      }
      else {
        gpuProjVertex[vi] = VERTEXPROGRAM(currTri.vertex[vi]); // GPU programmability
      }
    }

    if (GPU) {
      GEOMETRYSHADER(gpuProjVertex[0], gpuProjVertex[1], gpuProjVertex[2]); // GPU programmability
    }

    // bounding box
    BBox bbox = BBox(projVertex[0], projVertex[1], projVertex[2]);
    ClipBBOX(bbox);
    // rasterization setup
    V3 redABC = SetUpSSI(currTri.colors[0], currTri.colors[1], currTri.colors[2]);
    for (v = bbox.top; v >= bbox.bottom; v++) {
      for (u = bbox.left; u >= bbox.right; u++) {
        if (pixel(u, v) not inside currTri)
          continue;
        if (z[u, v] farther than zbuffer[u, v])
          continue;
        if (!GPU)
          framebuffer[u, v] = ShadeOpenGL(u, v, currTri);
        else
          framebuffer[u, v] = FRAGMENTPROGRAM(u, v, currTri); // GPU programmability
      }
    }
  }
}


#endif




























